# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

import io

import pytest
from rdflib import Graph

from invenio_vocabularies.contrib.subjects.gemet.datastreams import (
    GEMETSubjectsTransformer,
)
from invenio_vocabularies.datastreams.datastreams import StreamEntry
from invenio_vocabularies.datastreams.readers import RDFReader

XML_DATA = bytes(
    """<?xml version="1.0" encoding="UTF-8"?>
    <rdf:RDF xmlns:skos="http://www.w3.org/2004/02/skos/core#"
         xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#"
         xmlns:dcterms="http://purl.org/dc/terms/"
         xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
         xmlns:xsd="http://www.w3.org/2001/XMLSchema#"
         xml:base="http://www.eionet.europa.eu/gemet/">
        <skos:Concept rdf:about="concept/10008">
            <skos:inScheme rdf:resource="http://www.eionet.europa.eu/gemet/gemetThesaurus"/>
            <skos:prefLabel xml:lang="en">Consumer product</skos:prefLabel>
            <skos:prefLabel xml:lang="ar">منتج استهلاكي</skos:prefLabel>
            <skos:broader rdf:resource="concept/6660"/>
            <skos:memberOf rdf:resource="group/10112"/>
            <skos:memberOf rdf:resource="theme/27"/>
            <skos:memberOf rdf:resource="theme/34"/>
        </skos:Concept>
        <rdf:Description rdf:about="theme/27">
            <skos:member rdf:resource="concept/10008"/>
        </rdf:Description>
        <rdf:Description rdf:about="theme/34">
            <skos:member rdf:resource="concept/10008"/>
        </rdf:Description>
        <rdf:Description rdf:about="group/10112">
            <skos:member rdf:resource="concept/10008"/>
        </rdf:Description>
    </rdf:RDF>""",
    encoding="utf-8",
)


@pytest.fixture(scope="module")
def expected_from_rdf():
    return [
        {
            "id": "gemet:concept/10008",
            "scheme": "GEMET",
            "subject": "Consumer product",
            "title": {
                "en": "Consumer product",
                "ar": "منتج استهلاكي",
            },
            "props": {
                "parents": "gemet:concept/6660",
                "groups": ["http://www.eionet.europa.eu/gemet/group/10112"],
                "themes": [
                    "http://www.eionet.europa.eu/gemet/theme/27",
                    "http://www.eionet.europa.eu/gemet/theme/34",
                ],
            },
            "identifiers": [
                {
                    "scheme": "url",
                    "identifier": "http://www.eionet.europa.eu/gemet/concept/10008",
                }
            ],
        }
    ]


def test_gemet_concept_transformer_pref_label(expected_from_rdf):
    reader = RDFReader()
    rdf_data = io.BytesIO(XML_DATA)
    rdf_graph = Graph()
    rdf_graph.parse(rdf_data, format="xml")
    stream_entries = list(reader._iter(rdf_graph))
    assert len(stream_entries) > 0
    transformer = GEMETSubjectsTransformer()
    result = []
    for entry in stream_entries:
        entry = transformer.apply(StreamEntry(entry)).entry
        result.append(entry)
    assert expected_from_rdf == result



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#

"""Vocabulary generators."""

from invenio_access import any_user, authenticated_user
from invenio_records_permissions.generators import ConditionalGenerator
from invenio_search.engine import dsl


class IfTags(ConditionalGenerator):
    """Generator to filter based on tags.

    This generator will filter out records based on the tags field.
    """

    def __init__(self, tags, then_, else_):
        """Constructor."""
        self.tags = tags or []
        super().__init__(then_, else_)

    def _condition(self, record=None, **kwargs):
        """Check if the record has the tags."""
        return any(tag in record.get("tags", []) for tag in self.tags)

    def query_filter(self, **kwargs):
        """Search based on configured tags."""
        must_not_clauses = [dsl.Q("terms", tags=self.tags)]
        return dsl.Q(
            "bool",
            must_not=must_not_clauses,
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary models."""

from invenio_db import db
from invenio_records.models import RecordMetadataBase


class VocabularyType(db.Model):
    """Vocabulary type model."""

    __tablename__ = "vocabularies_types"

    id = db.Column(db.String(255), primary_key=True)
    pid_type = db.Column(db.String(255), unique=True)

    @classmethod
    def create(cls, **data):
        """Create a new vocabulary type."""
        with db.session.begin_nested():
            obj = cls(**data)
            db.session.add(obj)
        return obj

    @classmethod
    def dump_obj(cls, field, record, obj):
        """Serializer the object into a record."""
        record[field.attr_name] = {
            "id": obj.id,
            "pid_type": obj.pid_type,
        }

    @classmethod
    def load_obj(cls, field, record):
        """Deserializer the object from a record."""
        data = record.get(field.attr_name)
        if data:
            obj = cls(
                id=data.get("id"),
                pid_type=data.get("pid_type"),
            )
            return obj
        return None


class VocabularyMetadata(db.Model, RecordMetadataBase):
    """Model for vocabulary metadata."""

    __tablename__ = "vocabularies_metadata"


class VocabularyScheme(db.Model):
    """Vocabulary scheme model.

    This table stores the metadata for schemes (subtypes) of VocabularyType
    or separate specific vocabularies.

    It is only used to store metadata for subject schemes for now.
    But we might store affiliations's schemes or other schemes later.
    """

    __tablename__ = "vocabularies_schemes"

    id = db.Column(db.String(255), primary_key=True)
    # This is e.g. `subjects`, 'affiliations', ...
    parent_id = db.Column(db.String(255), primary_key=True)
    name = db.Column(db.String(255))
    uri = db.Column(db.String(255))
    # Any extra metadata is added as columns.

    @classmethod
    def create(cls, **data):
        """Create a new vocabulary subtype."""
        banned = [",", ":"]
        for b in banned:
            assert b not in data["id"], f"No '{b}' allowed in VocabularyScheme.id"

        with db.session.begin_nested():
            obj = cls(**data)
            db.session.add(obj)
        return obj



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Names data streams tests."""

import os
import tempfile
from copy import deepcopy
from unittest.mock import patch

import pytest
from invenio_access.permissions import system_identity
from invenio_records_resources.proxies import current_service_registry

from invenio_vocabularies.contrib.names.api import Name
from invenio_vocabularies.contrib.names.datastreams import (
    NamesServiceWriter,
    OrcidHTTPReader,
    OrcidTransformer,
)
from invenio_vocabularies.datastreams import StreamEntry
from invenio_vocabularies.datastreams.errors import TransformerError, WriterError
from invenio_vocabularies.datastreams.transformers import XMLTransformer


@pytest.fixture(scope="function")
def name_full_data():
    """Full name data."""
    return {
        "id": "0000-0001-8135-3489",
        "given_name": "Lars Holm",
        "family_name": "Nielsen",
        "name": "Nielsen, Lars Holm",
        "identifiers": [
            {
                "scheme": "orcid",
                "identifier": "0000-0001-8135-3489",  # normalized after create
            }
        ],
        "affiliations": [{"name": "CERN"}],
    }


# NOTE: This is a simplified version of the original XML data from ORCiD. Sections like
#       works, funding, educations, etc. and attributes have been removed.
XML_ENTRY_DATA = b"""
<record:record>
    <common:orcid-identifier>
        <common:uri>https://orcid.org/0000-0001-8135-3489</common:uri>
        <common:path>0000-0001-8135-3489</common:path>
        <common:host>orcid.org</common:host>
    </common:orcid-identifier>
    <person:person>
        <person:name>
            <personal-details:given-names>Lars Holm</personal-details:given-names>
            <personal-details:family-name>Nielsen</personal-details:family-name>
        </person:name>
    </person:person>
    <activities:activities-summary>
        <activities:employments>
            <activities:affiliation-group>
                <employment:employment-summary>
                    <common:start-date>
                        <common:year>2012</common:year>
                        <common:month>03</common:month>
                        <common:day>16</common:day>
                    </common:start-date>
                    <common:organization>
                        <common:name>CERN</common:name>
                        <common:disambiguated-organization>
                            <common:disambiguated-organization-identifier>https://ror.org/01ggx4157
</common:disambiguated-organization-identifier>
                            <common:disambiguation-source>ROR</common:disambiguation-source>
                        </common:disambiguated-organization>
                    </common:organization>
                </employment:employment-summary>
            </activities:affiliation-group>
            <activities:affiliation-group>
                <employment:employment-summary>
                    <common:start-date>
                        <common:year>2024</common:year>
                        <common:month>01</common:month>
                    </common:start-date>
                    <common:organization>
                        <common:name>ACME Inc.</common:name>
                        <common:disambiguated-organization>
                            <common:disambiguated-organization-identifier>grid.123456.z</common:disambiguated-organization-identifier>
                            <common:disambiguation-source>GRID</common:disambiguation-source>
                        </common:disambiguated-organization>
                    </common:organization>
                </employment:employment-summary>
            </activities:affiliation-group>
            <activities:affiliation-group>
                <employment:employment-summary >
                    <common:start-date>
                        <common:year>2007</common:year>
                        <common:month>08</common:month>
                    </common:start-date>
                    <common:end-date>
                        <common:year>2012</common:year>
                        <common:month>03</common:month>
                    </common:end-date>
                    <common:organization>
                        <common:name>European Southern Observatory</common:name>
                        <common:disambiguated-organization>
                            <common:disambiguated-organization-identifier>grid.424907.c</common:disambiguated-organization-identifier>
                            <common:disambiguation-source>GRID</common:disambiguation-source>
                        </common:disambiguated-organization>
                    </common:organization>
                </employment:employment-summary>
            </activities:affiliation-group>
        </activities:employments>
    </activities:activities-summary>
</record:record>
"""

XML_ENTRY_DATA_SINGLE_EMPLOYMENT = b"""
<record:record>
    <common:orcid-identifier>
        <common:uri>https://orcid.org/0000-0001-8135-3489</common:uri>
        <common:path>0000-0001-8135-3489</common:path>
        <common:host>orcid.org</common:host>
    </common:orcid-identifier>
    <person:person>
        <person:name>
            <personal-details:given-names>Lars Holm</personal-details:given-names>
            <personal-details:family-name>Nielsen</personal-details:family-name>
        </person:name>
    </person:person>
    <activities:activities-summary>
        <activities:employments>
            <activities:affiliation-group>
                <employment:employment-summary>
                    <common:start-date>
                        <common:year>2012</common:year>
                        <common:month>03</common:month>
                        <common:day>16</common:day>
                    </common:start-date>
                    <common:organization>
                        <common:name>CERN</common:name>
                        <common:disambiguated-organization>
                            <common:disambiguated-organization-identifier>https://ror.org/01ggx4157
</common:disambiguated-organization-identifier>
                            <common:disambiguation-source>ROR</common:disambiguation-source>
                        </common:disambiguated-organization>
                    </common:organization>
                </employment:employment-summary>
            </activities:affiliation-group>
        </activities:employments>
    </activities:activities-summary>
</record:record>
"""


NAMES_TEST = {
    "Lars Holm, Nielsen": True,
    "Lars Nielsen:)": False,
    "Lars Nielsen-H": True,
    "Lars Nielsen 👍": False,
    "Lars Nielsen (CERN)": True,
}


@pytest.fixture(scope="function")
def orcid_data():
    base = {
        "orcid-identifier": {
            "uri": "https://orcid.org/0000-0001-8135-3489",
            "path": "0000-0001-8135-3489",
            "host": "orcid.org",
        },
        "person": {
            "name": {"given-names": "Lars Holm", "family-name": "Nielsen"},
        },
    }

    employments = [
        {
            "employment-summary": {
                "organization": {
                    "name": "CERN",
                    "disambiguated-organization": {
                        "disambiguated-organization-identifier": "https://ror.org/01ggx4157",
                        "disambiguation-source": "ROR",
                    },
                },
                "start-date": {"year": "2012", "month": "03", "day": "16"},
            }
        },
        {
            "employment-summary": {
                "organization": {
                    "name": "ACME Inc.",
                    "disambiguated-organization": {
                        "disambiguated-organization-identifier": "grid.123456.z",
                        "disambiguation-source": "GRID",
                    },
                },
                "start-date": {"year": "2024", "month": "01"},
            }
        },
        {
            "employment-summary": {
                "organization": {
                    "name": "European Southern Observatory",
                    "disambiguated-organization": {
                        "disambiguated-organization-identifier": "grid.424907.c",
                        "disambiguation-source": "GRID",
                    },
                },
                "start-date": {"year": "2007", "month": "08"},
                "end-date": {"year": "2012", "month": "03"},
            },
        },
    ]

    return {
        "xml": {
            "multi_employment": XML_ENTRY_DATA,
            "single_employment": XML_ENTRY_DATA_SINGLE_EMPLOYMENT,
        },
        "json": {
            "multi_employment": {
                **base,
                "activities-summary": {
                    "employments": {"affiliation-group": employments},
                },
            },
            "single_employment": {
                **base,
                "activities-summary": {
                    # NOTE: Because the XML data has only one employment, the
                    # transformer will not create a list of employments, but instead a
                    # single employment object.
                    "employments": {"affiliation-group": employments[0]}
                },
            },
            "duplicate_affiliations": {
                **base,
                "activities-summary": {
                    "employments": {"affiliation-group": employments + employments},
                },
            },
        },
    }


@pytest.fixture(scope="module")
def expected_from_xml():
    base = {
        "id": "0000-0001-8135-3489",
        "given_name": "Lars Holm",
        "family_name": "Nielsen",
        "identifiers": [{"scheme": "orcid", "identifier": "0000-0001-8135-3489"}],
    }

    return {
        "multi_employment": {
            **base,
            "affiliations": [
                {"id": "01ggx4157", "name": "CERN"},
                # NOTE: GRID identifiers do not result in linked affiliations
                {"name": "ACME Inc."},
                # NOTE: terminated employments are not included in the affiliations
            ],
        },
        "single_employment": {
            **base,
            "affiliations": [{"id": "01ggx4157", "name": "CERN"}],
        },
        "duplicate_affiliations": {
            **base,
            "affiliations": [
                # Affiliations are deduplicated
                {"id": "01ggx4157", "name": "CERN"},
                {"name": "ACME Inc."},
            ],
        },
    }


@pytest.fixture(scope="function")
def org_ids_mapping_file_config(app):
    """ORCiD organization IDs mappings CSV file."""
    fout = tempfile.NamedTemporaryFile(mode="w", delete=False)
    fout.write('"GRID","grid.123456.z","acme_inc_id"\n')
    fout.close()

    old_config = app.config.get("VOCABULARIES_ORCID_ORG_IDS_MAPPING_PATH")
    app.config["VOCABULARIES_ORCID_ORG_IDS_MAPPING_PATH"] = fout.name
    yield

    app.config["VOCABULARIES_ORCID_ORG_IDS_MAPPING_PATH"] = old_config
    os.unlink(fout.name)


def test_orcid_xml_transform(orcid_data):
    """Test XML transformer with ORCiD record.

    NOTE: this might look somewhat "redundant", since we're again testing if the
    XMLTransformer works as expected. However, this is a more specific test
    that demonstrates how the transformer behaves with more complex XML data, that
    can have e.g. nested elements that can be arrays or objects.
    """
    xml_data = orcid_data["xml"]
    json_data = orcid_data["json"]

    for key, data in xml_data.items():
        transformer = XMLTransformer(root_element="record")
        transform_result = transformer.apply(StreamEntry(data)).entry
        assert transform_result == json_data[key]


def test_orcid_transformer(app, orcid_data, expected_from_xml):
    """Test ORCiD transformer data."""
    transformer = OrcidTransformer()
    input_data = orcid_data["json"]

    for key, data in input_data.items():
        assert expected_from_xml[key] == transformer.apply(StreamEntry(data)).entry, key


def test_orcid_transformer_org_ids_mapping_from_file(
    app,
    orcid_data,
    expected_from_xml,
    org_ids_mapping_file_config,
):
    """Test ORCiD transformer data with org IDs mapping file."""
    transformer = OrcidTransformer()
    input_data = orcid_data["json"]["multi_employment"]
    expected = deepcopy(expected_from_xml["multi_employment"])
    expected["affiliations"] = [
        {"id": "01ggx4157", "name": "CERN"},
        {"id": "acme_inc_id", "name": "ACME Inc."},
    ]

    assert expected == transformer.apply(StreamEntry(input_data)).entry


@pytest.mark.parametrize("name,is_valid_name", NAMES_TEST.items())
def test_orcid_transformer_name_filtering(orcid_data, name, is_valid_name):
    transformer = OrcidTransformer()
    val = deepcopy(orcid_data["json"]["multi_employment"])
    if is_valid_name:
        assert transformer.apply(StreamEntry(val)).entry
    else:
        with pytest.raises(TransformerError):
            val["person"]["name"]["given-names"] = name
            val["person"]["name"]["family-name"] = ""
            transformer.apply(StreamEntry(val))
        with pytest.raises(TransformerError):
            val["person"]["name"]["given-names"] = ""
            val["person"]["name"]["family-name"] = name
            transformer.apply(StreamEntry(val))


@pytest.mark.parametrize("name", NAMES_TEST.keys())
def test_orcid_transformer_different_names_no_regex(orcid_data, name):
    transformer = OrcidTransformer(names_exclude_regex=None)
    val = deepcopy(orcid_data["json"]["multi_employment"])
    val["person"]["name"]["given-names"] = name
    val["person"]["name"]["family-name"] = ""
    assert transformer.apply(StreamEntry(val)).entry


class MockResponse:
    content = XML_ENTRY_DATA
    status_code = 200


@patch("requests.get", side_effect=lambda url, headers: MockResponse())
def test_orcid_http_reader(_):
    reader = OrcidHTTPReader(id="0000-0001-8135-3489")
    results = []
    for entry in reader.read():
        results.append(entry)

    assert len(results) == 1
    assert XML_ENTRY_DATA == results[0]


def test_names_service_writer_create(app, db, search_clear, name_full_data):
    writer = NamesServiceWriter()
    record = writer.write(StreamEntry(name_full_data))
    record = record.entry.to_dict()

    assert dict(record, **name_full_data) == record


def test_names_service_writer_duplicate(app, db, search_clear, name_full_data):
    writer = NamesServiceWriter()
    _ = writer.write(stream_entry=StreamEntry(name_full_data))
    Name.index.refresh()  # refresh index to make changes live
    with pytest.raises(WriterError) as err:
        writer.write(stream_entry=StreamEntry(name_full_data))

    expected_error = [f"Vocabulary entry already exists: {name_full_data}"]
    assert expected_error in err.value.args


def test_names_service_writer_update_existing(app, db, search_clear, name_full_data):
    # create vocabulary
    writer = NamesServiceWriter(update=True)
    name = writer.write(stream_entry=StreamEntry(name_full_data))
    Name.index.refresh()  # refresh index to make changes live
    # update vocabulary
    updated_name = deepcopy(name_full_data)
    updated_name["given_name"] = "Pablo"
    updated_name["family_name"] = "Panero"
    updated_name["name"] = "Panero, Pablo"
    # check changes vocabulary
    _ = writer.write(stream_entry=StreamEntry(updated_name))
    service = current_service_registry.get("names")
    record = service.read(system_identity, name.entry.id)
    record = record.to_dict()

    # needed while the writer resolves from ES
    assert _.entry.id == name.entry.id
    assert dict(record, **updated_name) == record


def test_names_service_writer_update_non_existing(
    app, db, search_clear, name_full_data
):
    # vocabulary item not created, call update directly
    updated_name = deepcopy(name_full_data)
    updated_name
    updated_name["given_name"] = "Pablo"
    updated_name["family_name"] = "Panero"
    updated_name["name"] = "Panero, Pablo"
    # check changes vocabulary
    writer = NamesServiceWriter(update=True)
    name = writer.write(stream_entry=StreamEntry(updated_name))
    service = current_service_registry.get("names")
    record = service.read(system_identity, name.entry.id)
    record = record.to_dict()

    assert dict(record, **updated_name) == record



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabularies test config."""

import pytest
from invenio_indexer.api import RecordIndexer

from invenio_vocabularies.records.api import Vocabulary


@pytest.fixture()
def indexer():
    """Indexer instance with correct Record class."""
    return RecordIndexer(
        record_cls=Vocabulary,
        record_to_index=lambda r: (r._record.index._name, "_doc"),
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""GEMET subjects datastreams, readers, transformers, and writers."""

from invenio_vocabularies.datastreams.transformers import RDFTransformer

from ..config import gemet_file_url

# Available with the "rdf" extra
try:
    import rdflib
except ImportError:
    rdflib = None


class GEMETSubjectsTransformer(RDFTransformer):
    """
    Transformer class to convert GEMET RDF data to a dictionary format.

    Input:
        - Relevant fields:
            - `skos:prefLabel`: Preferred labels with language codes.
            - `skos:broader`: References to broader concepts (parent concepts).
            - `skos:memberOf`: References to groups or themes the concept belongs to.

    Output:
        - A dictionary with the following structure:
            {
                "id": "gemet:concept/10008",  # GEMET-specific concept ID (skos:Concept).
                "scheme": "GEMET",  # The scheme name indicating this is a GEMET concept.
                "subject": "Consumer product",  # The subject label (first preferred label in English, skos:prefLabel).
                "title": {
                    "en": "Consumer product",  # English label for the concept (skos:prefLabel).
                    "ar": "منتج استهلاكي"  # Arabic label for the concept (skos:prefLabel).
                },
                "props": {
                    "parents": "gemet:concept/6660",  # The parent concept (skos:broader), identified by its GEMET Concept ID.
                    "groups": ["http://www.eionet.europa.eu/gemet/group/10112"],  # Group the concept belongs to (skos:memberOf)(skos:prefLabel).
                    "themes": [
                        "http://www.eionet.europa.eu/gemet/theme/27",  # Theme the concept belongs to (skos:memberOf)(rdfs:label).
                    ]
                },
                "identifiers": [
                    {
                        "scheme": "url",  # Type of identifier (URL).
                        "identifier": "http://www.eionet.europa.eu/gemet/concept/10008"  # URI of the concept (rdf:about).
                    }
                ]
            }
    """

    def _get_parent_notation(self, broader, rdf_graph):
        """Extract parent notation from GEMET URI."""
        return "/".join(broader.split("/")[-2:])

    def _get_groups_and_themes(self, subject, rdf_graph):
        """Extract groups and themes for a subject."""
        groups = []
        themes = []

        for relation in rdf_graph.subjects(
            predicate=self.skos_core.member, object=subject
        ):
            relation_uri = str(relation)
            relation_label = None

            # If the relation is a group, check for skos:prefLabel
            if "group" in relation_uri:
                labels = rdf_graph.objects(
                    subject=relation, predicate=self.skos_core.prefLabel
                )
                relation_label = next(
                    (str(label) for label in labels if label.language == "en"), None
                )
                groups.append(relation_uri)

            # If the relation is a theme, check for rdfs:label
            elif "theme" in relation_uri:
                labels = rdf_graph.objects(
                    subject=relation, predicate=rdflib.RDFS.label
                )
                relation_label = next(
                    (str(label) for label in labels if label.language == "en"), None
                )
                themes.append(relation_uri)

        return groups, themes

    def _transform_entry(self, subject, rdf_graph):
        """Transform an entry to the required dictionary format."""
        concept_number = "/".join(subject.split("/")[-2:])
        id = f"gemet:{concept_number}" if concept_number else None
        labels = self._get_labels(subject, rdf_graph)
        parents = ",".join(
            f"gemet:{n}" for n in reversed(self._find_parents(subject, rdf_graph)) if n
        )
        identifiers = [{"scheme": "url", "identifier": str(subject)}]
        groups, themes = self._get_groups_and_themes(subject, rdf_graph)

        props = {"parents": parents} if parents else {}

        if groups:
            props["groups"] = groups
        if themes:
            props["themes"] = themes

        return {
            "id": id,
            "scheme": "GEMET",
            "subject": labels.get("en", "").capitalize(),
            "title": labels,
            "props": props,
            "identifiers": self._get_identifiers(subject),
        }


# Configuration for datastream

VOCABULARIES_DATASTREAM_TRANSFORMERS = {"gemet-transformer": GEMETSubjectsTransformer}

DATASTREAM_CONFIG = {
    "readers": [
        {
            "type": "http",
            "args": {
                "origin": gemet_file_url,
            },
        },
        {"type": "gzip"},
        {"type": "rdf"},
    ],
    "transformers": [{"type": "gemet-transformer"}],
    "writers": [{"args": {"writer": {"type": "subjects-service"}}, "type": "async"}],
}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Basic tests for the mock module to ensure it works properly."""

from functools import partial

import pytest
from invenio_indexer.api import RecordIndexer
from invenio_search import current_search_client
from mock_module.api import Record


#
# Fixtures
#
@pytest.fixture()
def mock_record(app, db, example_record):
    """An example mock record."""
    return Record.create({}, metadata={"title": "Test", "languages": [{"id": "eng"}]})


@pytest.fixture()
def mock_indexer():
    """Get an indexer for mock records."""
    return RecordIndexer(
        record_cls=Record,
        record_to_index=lambda r: (r.__class__.index._name, "_doc"),
    )


@pytest.fixture()
def mock_search():
    """Get a search client."""
    return partial(current_search_client.get, Record.index._name)


#
# Tests
#
def test_mock_record(mock_record):
    """Basic smoke test."""
    assert mock_record.schema
    assert mock_record.pid


def test_linked_record(mock_record, example_record):
    """Linked record fetching."""
    # Dereference the linked language record
    lang_record = list(mock_record.relations.languages())[0]
    assert lang_record == example_record


def test_dereferencing(mock_record):
    """Record dereferencing."""
    # Dereference the linked language record
    mock_record.relations.languages.dereference()
    deferenced_lang_record = mock_record.metadata["languages"][0]
    # Test that only part of the language record is denormalised.
    assert sorted(deferenced_lang_record.keys()) == ["@v", "id", "title"]


def test_dumping(mock_record, example_record):
    """Record schema validation."""
    # Create a record linked to a language record.
    lang = mock_record.dumps()["metadata"]["languages"][0]
    assert lang == {
        "id": "eng",
        "title": {"da": "Engelsk", "en": "English"},
        "@v": str(example_record.id) + "::" + str(example_record.revision_id),
    }


def test_indexing(mock_record, mock_indexer, mock_search, example_record):
    # Index document in ES
    assert mock_indexer.index(mock_record)["result"] == "created"

    # Retrieve document from ES and load the source
    data = mock_search(id=mock_record.id)
    record = Record.loads(data["_source"])

    # Getting the language records should work:
    lang_record = list(record.relations.languages())[0]
    assert lang_record == example_record

    # Dereferencing also works
    record.relations.languages.dereference()
    deferenced_lang_record = mock_record.metadata["languages"][0]
    assert sorted(deferenced_lang_record.keys()) == ["@v", "id", "title"]



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test the vocabulary service."""

import arrow
import pytest
from invenio_cache import current_cache
from invenio_pidstore.errors import PIDAlreadyExists, PIDDeletedError
from marshmallow.exceptions import ValidationError
from sqlalchemy.exc import IntegrityError

from invenio_vocabularies.records.api import Vocabulary, VocabularyType


#
# Fixtures
#
@pytest.fixture()
def lic_type(db):
    """Get a language vocabulary type."""
    return VocabularyType.create(id="licenses", pid_type="lic")


#
# Tests
#
def test_simple_flow(lang_type, lic_type, lang_data, service, identity):
    """Test a simple vocabulary service flow."""
    # Create it
    item = service.create(identity, lang_data)
    id_ = item.id

    assert item.id == lang_data["id"]
    for k, v in lang_data.items():
        assert item.data[k] == v

    # Read it
    read_item = service.read(identity, ("languages", "eng"))
    assert item.id == read_item.id
    assert item.data == read_item.data
    assert read_item.data["type"] == "languages"

    # Refresh index to make changes live.
    Vocabulary.index.refresh()

    # Search it
    res = service.search(identity, type="languages", q=f"id:{id_}", size=25, page=1)
    assert res.total == 1
    assert list(res.hits)[0] == read_item.data

    # Search another type
    res = service.search(identity, type="licenses", q=f"id:{id_}", size=25, page=1)
    assert res.total == 0

    # Update it
    data = read_item.data
    data["title"]["en"] = "New title"
    update_item = service.update(identity, ("languages", id_), data)
    assert item.id == update_item.id
    assert update_item["title"]["en"] == "New title"

    # Delete it
    assert service.delete(identity, ("languages", id_))

    # Refresh to make changes live
    Vocabulary.index.refresh()

    # Fail to retrieve it
    # - db
    pytest.raises(PIDDeletedError, service.read, identity, ("languages", id_))
    # - search
    res = service.search(identity, type="languages", q=f"id:{id_}", size=25, page=1)
    assert res.total == 0


def test_pid_already_registered(lang_type, lang_data, service, identity):
    """Recreating a record with same id should fail."""
    service.create(identity, lang_data)
    pytest.raises(PIDAlreadyExists, service.create, identity, lang_data)


def test_extra_fields(lang_data2, service, identity):
    """Extra fields in data should fail."""
    lang_data2["invalid"] = 1
    pytest.raises(ValidationError, service.create, identity, lang_data2)


def test_missing_or_invalid_type(lang_data2, service, identity):
    """A missing type should raise validation error."""
    # No type specified
    del lang_data2["type"]
    pytest.raises(ValidationError, service.create, identity, lang_data2)

    # Invalid value data types
    for val in [
        1,
        {"id": "languages"},
    ]:
        lang_data2["type"] = val
        pytest.raises(ValidationError, service.create, identity, lang_data2)

    # Non-existing type
    lang_data2["type"] = "invalid"
    pytest.raises(ValidationError, service.create, identity, lang_data2)


def test_update(lang_type, lang_data2, service, identity):
    """Removing keys should work."""
    lang_data2["id"] = "dan"
    item = service.create(identity, lang_data2)
    id_ = item.id

    data = item.data
    # Unset a subkey
    del data["title"]["en"]
    # Unset a top-level key
    del data["icon"]
    update_item = service.update(identity, ("languages", id_), data)
    # Ensure they really got cleared.
    assert "en" not in update_item.data["title"]
    assert "icon" not in update_item.data


def test_missing_or_invalid_type(lang_data2, service, identity):
    """Custom props should only accept strings."""
    # Invalid data types
    for v in [1, {"id": "test"}]:
        lang_data2["props"]["newkey"] = v
        pytest.raises(ValidationError, service.create, identity, lang_data2)


def test_read_all_no_cache(lang_type, lang_data_many, service, identity, cache):
    """read_all method should return all languages created in this scope."""
    items = service.read_all(identity, fields=["id"], type="languages", cache=False)
    assert set(lang_data_many).issubset(set([i["id"] for i in items]))
    cached = current_cache.get("id")
    assert not cached


def test_read_all_cache(lang_type, lang_data_many, service, identity, cache):
    """read_all method should return all languages created in this scope."""
    items = service.read_all(identity, fields=["id"], type="languages", cache=True)
    assert set(lang_data_many).issubset(set([i["id"] for i in items]))
    cached = current_cache.get("languages__id")
    assert cached is not None


def test_read_many(lang_type, lang_data_many, service, identity, search_clear):
    """read_many method should return all requested languages."""
    ids_ = ["fr", "tr", "es"]
    items = service.read_many(identity, type="languages", ids=ids_, fields=[])
    item_ids = [i.id for i in items._results]
    assert {"fr", "tr", "es"} == set(item_ids)


def test_indexed_at_query(app, db, service, identity, lang_type, lang_data):
    before = arrow.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%f")
    _ = service.create(identity, lang_data)
    now = arrow.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%f")
    Vocabulary.index.refresh()

    # there is previous to before
    res = service.search(
        identity,
        q=f"indexed_at:[* TO {before}]",
        type="languages",
    )
    assert res.total == 0

    # there is previous to now
    res = service.search(
        identity,
        q=f"indexed_at:[* TO {now}]",
        type="languages",
    )
    assert res.total == 1



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test the names vocabulary permissions."""

import pytest
from flask_principal import Identity
from invenio_access.permissions import any_user, authenticated_user
from invenio_records_resources.services.errors import (
    PermissionDeniedError,
    RecordPermissionDeniedError,
)


#
# Fixtures
#
@pytest.fixture()
def anyuser_idty():
    """Simple identity to interact with the service."""
    identity = Identity(1)
    identity.provides.add(any_user)
    return identity


@pytest.fixture()
def authenticated_user_idty():
    """Authenticated identity to interact with the service."""
    identity = Identity(2)
    identity.provides.add(authenticated_user)
    return identity


def test_non_searchable_tag(
    app,
    service,
    identity,
    non_searchable_name_data,
    anyuser_idty,
    authenticated_user_idty,
    example_affiliation,
    superuser_identity,
    indexer,
):
    """Test that unlisted tags are not returned in search results."""
    # Service
    assert service.id == "names"
    assert service.config.indexer_queue_name == "names"
    # Create it
    item = service.create(identity, non_searchable_name_data)
    id_ = item.id

    # Index document in ES
    assert indexer.refresh()

    with pytest.raises(PermissionDeniedError):
        res = service.search(anyuser_idty, type="names", q=f"id:{id_}", size=25, page=1)

    # Search - only searchable values should be returned
    res = service.search(
        authenticated_user_idty, type="names", q=f"id:{id_}", size=25, page=1
    )
    assert res.total == 0
    # Anyuser should not be able to read the record
    with pytest.raises(RecordPermissionDeniedError):
        service.read(anyuser_idty, id_)

    # Authenticated user should not be able to read the record
    with pytest.raises(RecordPermissionDeniedError):
        test = service.read(authenticated_user_idty, id_)

    # Admins should be able to see the unlisted tags
    res = service.search(
        superuser_identity, type="names", q=f"id:{id_}", size=25, page=1
    )
    assert res.total == 1

    item = service.read(superuser_identity, id_)

    assert item["id"] == id_



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2024 CERN.
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Invenio vocabularies views."""

from flask import Blueprint

blueprint = Blueprint(
    "invenio_vocabularies_ext", __name__, template_folder="./templates"
)


def create_blueprint_from_app(app):
    """Create app blueprint."""
    return app.extensions["invenio-vocabularies"].resource.as_blueprint()


def create_affiliations_blueprint_from_app(app):
    """Create app blueprint."""
    return app.extensions["invenio-vocabularies"].affiliations_resource.as_blueprint()


def create_awards_blueprint_from_app(app):
    """Create app blueprint."""
    return app.extensions["invenio-vocabularies"].awards_resource.as_blueprint()


def create_funders_blueprint_from_app(app):
    """Create app blueprint."""
    return app.extensions["invenio-vocabularies"].funders_resource.as_blueprint()


def create_names_blueprint_from_app(app):
    """Create app blueprint."""
    return app.extensions["invenio-vocabularies"].names_resource.as_blueprint()


def create_subjects_blueprint_from_app(app):
    """Create app blueprint."""
    return app.extensions["invenio-vocabularies"].subjects_resource.as_blueprint()


def create_list_blueprint_from_app(app):
    """Create app blueprint."""
    return app.extensions[
        "invenio-vocabularies"
    ].vocabulary_admin_resource.as_blueprint()



#
# This file is part of Invenio.
# Copyright (C) 2024 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Add internal name ID."""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "3ba812d80559"
down_revision = "55a700f897b6"
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    op.add_column(
        "name_metadata", sa.Column("internal_id", sa.String(length=255), nullable=True)
    )
    op.create_unique_constraint(
        op.f("uq_name_metadata_internal_id"), "name_metadata", ["internal_id"]
    )


def downgrade():
    """Downgrade database."""
    op.drop_constraint(
        op.f("uq_name_metadata_internal_id"), "name_metadata", type_="unique"
    )
    op.drop_column("name_metadata", "internal_id")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Data Streams transformers tests."""

import pytest

from invenio_vocabularies.datastreams import StreamEntry
from invenio_vocabularies.datastreams.errors import TransformerError
from invenio_vocabularies.datastreams.transformers import XMLTransformer


@pytest.fixture(scope="module")
def expected_from_xml():
    return {
        "top_level_field": "top-level single value",
        "top_level_object_field": {
            "some": "value",
            "another": "value too",
            "nested_array_field": {
                "@array_attr": "value",
                "array_element_object": [
                    {
                        "@obj_attr": "first",
                        "element_foo": "value1",
                        "element_bar": "value1",
                    },
                    {
                        "@obj_attr": "second",
                        "element_foo": "value2",
                        "element_bar": "value2",
                        "element_qux": "value2",
                    },
                ],
            },
        },
    }


def test_xml_transformer(expected_from_xml):
    bytes_xml_entry = StreamEntry(
        b"""
        <?xml version="1.0" encoding="UTF-8" standalone="yes"?>
        <ns:top_level_field>top-level single value</ns:top_level_field>
        <ns:top_level_object_field>
            <ns:some>value</ns:some>
            <ns:another>value too</ns:another>
            <ns:nested_array_field array_attr="value">
                <ns:array_element_object obj_attr="first">
                    <ns:element_foo>value1</ns:element_foo>
                    <ns:element_bar>value1</ns:element_bar>
                </ns:array_element_object>
                <ns:array_element_object obj_attr="second">
                    <ns:element_foo>value2</ns:element_foo>
                    <ns:element_bar>value2</ns:element_bar>
                    <ns:element_qux>value2</ns:element_qux>
                </ns:array_element_object>
            </ns:nested_array_field array_attr="value">
        </ns:top_level_object_field>
        """
    )

    transformer = XMLTransformer()
    assert expected_from_xml == transformer.apply(bytes_xml_entry).entry


def test_bad_xml_transformer():
    bytes_xml_entry = StreamEntry(
        b"""
        <?xml version="1.0" encoding="UTF-8" standalone="yes"?>
        <ns:top_level_field>top-level single value</ns:top_level_field>
        <ns:top_level_object_field>
            <ns:some>value</ns:some>
            <ns:another>value too</ns:another>
            <ns:nested_array_field array_attr="value">
                <ns:array_element_object obj_attr="first">
                    <ns:element_foo>value1</ns:element_foo>
                    <ns:element_bar>value1</ns:element_bar>
                </ns:array_element_object>
                <ns:array_element_object obj_attr="second">
                    <ns:element_foo>value2</ns:element_foo>
                    <ns:element_bar>value2</ns:element_bar>
                    <ns:element_qux>value2</ns:element_qux>
                </ns:array_element_object>
            </ns:nested_array_field array_attr="value">
        </ns:top_level_object_field>
        """
    )

    transformer = XMLTransformer(root_element="field_two")

    with pytest.raises(TransformerError):
        transformer.apply(bytes_xml_entry)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Names schema."""

from functools import partial

from invenio_i18n import lazy_gettext as _
from marshmallow import (
    EXCLUDE,
    ValidationError,
    fields,
    post_dump,
    post_load,
    validates_schema,
)
from marshmallow_utils.fields import IdentifierSet, SanitizedUnicode
from marshmallow_utils.schemas import IdentifierSchema

from ...services.schema import BaseVocabularySchema, ModePIDFieldVocabularyMixin
from ..affiliations.schema import (
    AffiliationRelationSchema as BaseAffiliationRelationSchema,
)
from .config import names_schemes


class AffiliationRelationSchema(BaseAffiliationRelationSchema):
    """Affiliation relation schema."""

    acronym = SanitizedUnicode(dump_only=True)

    class Meta:
        """Meta class."""

        unknown = EXCLUDE


class NameSchema(BaseVocabularySchema, ModePIDFieldVocabularyMixin):
    """Service schema for names.

    Note that this vocabulary is very different from the base vocabulary
    so it does not inherit from it.
    """

    internal_id = fields.Str(allow_none=True)
    name = SanitizedUnicode()
    given_name = SanitizedUnicode()
    family_name = SanitizedUnicode()
    identifiers = IdentifierSet(
        fields.Nested(
            partial(
                IdentifierSchema,
                # It is intended to allow org schemes to be sent as personal
                # and viceversa. This is a trade off learnt from running
                # Zenodo in production.
                allowed_schemes=names_schemes,
            )
        )
    )
    affiliations = fields.List(fields.Nested(AffiliationRelationSchema))
    props = fields.Dict(keys=fields.Str(), values=fields.Raw())

    @validates_schema
    def validate_names(self, data, **kwargs):
        """Validate names."""
        can_compose = data.get("family_name") and data.get("given_name")
        name = data.get("name")
        if not can_compose and not name:
            messages = [
                _(
                    "A name or the family name together with the given name must be present."
                )
            ]
            raise ValidationError({"family_name": messages})

    @validates_schema
    def validate_affiliations(self, data, **kwargs):
        """Validate names."""
        affiliations = data.get("affiliations", [])
        seen_names = set()
        for affiliation in affiliations:
            name = affiliation.get("name")
            if not affiliation.get("id") and name:
                if name in seen_names:
                    messages = [_("Duplicated affiliations.")]
                    raise ValidationError({"affiliations": messages})
                else:
                    seen_names.add(name)

    @post_load
    def update_name(self, data, **kwargs):
        """Update names for person.

        Fill name from given_name and family_name if person.
        """
        family_name = data.get("family_name")
        given_name = data.get("given_name")
        if family_name and given_name:
            data["name"] = f"{family_name}, {given_name}"

        return data

    @post_dump
    def dump_name(self, data, **kwargs):
        """Dumps the name if not present in the serialized data."""
        name = data.get("name")
        if not name:
            family_name = data.get("family_name")
            given_name = data.get("given_name")
            if family_name and given_name:
                data["name"] = f"{family_name}, {given_name}"
            elif family_name:
                data["name"] = family_name
        return data



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""OpenAIRE-related Datastreams Readers/Writers/Transformers module."""

import io

import requests

from invenio_vocabularies.datastreams.errors import ReaderError
from invenio_vocabularies.datastreams.readers import BaseReader


class OpenAIREHTTPReader(BaseReader):
    """OpenAIRE HTTP Reader returning an in-memory binary stream of the latest OpenAIRE Graph Dataset tar file of a given type."""

    def __init__(self, origin=None, mode="r", tar_href=None, *args, **kwargs):
        """Constructor."""
        self.tar_href = tar_href
        super().__init__(origin, mode, *args, **kwargs)

    def _iter(self, fp, *args, **kwargs):
        raise NotImplementedError(
            "OpenAIREHTTPReader downloads one file and therefore does not iterate through items"
        )

    def read(self, item=None, *args, **kwargs):
        """Reads the latest OpenAIRE Graph Dataset tar file of a given type from Zenodo and yields an in-memory binary stream of it."""
        if item:
            raise NotImplementedError(
                "OpenAIREHTTPReader does not support being chained after another reader"
            )

        if self._origin == "full":
            # OpenAIRE Graph Dataset
            api_url = "https://zenodo.org/api/records/3516917"
        elif self._origin == "diff":
            # OpenAIRE Graph dataset: new collected projects
            api_url = "https://zenodo.org/api/records/6419021"
        else:
            raise ReaderError("The --origin option should be either 'full' or 'diff'")

        # Call the signposting `linkset+json` endpoint for the Concept DOI (i.e. latest version) of the OpenAIRE Graph Dataset.
        # See: https://github.com/inveniosoftware/rfcs/blob/master/rfcs/rdm-0071-signposting.md#provide-an-applicationlinksetjson-endpoint
        headers = {"Accept": "application/linkset+json"}
        api_resp = requests.get(api_url, headers=headers)
        api_resp.raise_for_status()

        # Extract the Landing page Link Set Object located as the first (index 0) item.
        landing_page_linkset = api_resp.json()["linkset"][0]

        # Extract the URL of the only tar file matching `tar_href` linked to the record.
        landing_page_matching_tar_items = [
            item
            for item in landing_page_linkset["item"]
            if item["type"] == "application/x-tar"
            and item["href"].endswith(self.tar_href)
        ]
        if len(landing_page_matching_tar_items) != 1:
            raise ReaderError(
                f"Expected 1 tar item matching {self.tar_href} but got {len(landing_page_matching_tar_items)}"
            )
        file_url = landing_page_matching_tar_items[0]["href"]

        # Download the matching tar file and fully load the response bytes content in memory.
        # The bytes content are then wrapped by a BytesIO to be file-like object (as required by `tarfile.open`).
        # Using directly `file_resp.raw` is not possible since `tarfile.open` requires the file-like object to be seekable.
        file_resp = requests.get(file_url)
        file_resp.raise_for_status()
        yield io.BytesIO(file_resp.content)


VOCABULARIES_DATASTREAM_READERS = {
    "openaire-http": OpenAIREHTTPReader,
}

VOCABULARIES_DATASTREAM_TRANSFORMERS = {}

VOCABULARIES_DATASTREAM_WRITERS = {}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 CERN.
# Copyright (C) 2024 California Institute of Technology.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Funders datastreams, transformers, writers and readers."""

from flask import current_app
from invenio_i18n import lazy_gettext as _

from ...datastreams.writers import ServiceWriter
from ..common.ror.datastreams import RORTransformer


class FundersServiceWriter(ServiceWriter):
    """Funders service writer."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        service_or_name = kwargs.pop("service_or_name", "funders")
        super().__init__(service_or_name=service_or_name, *args, **kwargs)

    def _entry_id(self, entry):
        """Get the id from an entry."""
        return entry["id"]


class FundersRORTransformer(RORTransformer):
    """Funders ROR Transformer."""

    def __init__(
        self, *args, vocab_schemes=None, funder_fundref_doi_prefix=None, **kwargs
    ):
        """Constructor."""
        if vocab_schemes is None:
            vocab_schemes = current_app.config.get("VOCABULARIES_FUNDER_SCHEMES")
        if funder_fundref_doi_prefix is None:
            funder_fundref_doi_prefix = current_app.config.get(
                "VOCABULARIES_FUNDER_DOI_PREFIX"
            )
        super().__init__(
            *args,
            vocab_schemes=vocab_schemes,
            funder_fundref_doi_prefix=funder_fundref_doi_prefix,
            **kwargs,
        )


VOCABULARIES_DATASTREAM_READERS = {}
"""Funders datastreams writers."""

VOCABULARIES_DATASTREAM_WRITERS = {
    "funders-service": FundersServiceWriter,
}
"""Funders datastreams writers."""


VOCABULARIES_DATASTREAM_TRANSFORMERS = {
    "ror-funders": FundersRORTransformer,
}
"""Funders datastreams transformers."""


DATASTREAM_CONFIG = {
    "readers": [
        {
            "type": "zip",
            "args": {
                "regex": "_schema_v2\\.json$",
            },
        },
        {"type": "json"},
    ],
    "transformers": [
        {
            "type": "ror-funders",
        },
    ],
    "writers": [
        {
            "type": "async",
            "args": {
                "writer": {
                    "type": "funders-service",
                }
            },
        }
    ],
}
"""Data Stream configuration.

An origin is required for the reader.
"""



#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Drop unique constraint from internal id."""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "af2457652217"
down_revision = "3ba812d80559"
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    op.drop_constraint("uq_name_metadata_internal_id", "name_metadata", type_="unique")
    op.create_index(
        op.f("ix_name_metadata_internal_id"),
        "name_metadata",
        ["internal_id"],
        unique=False,
    )


def downgrade():
    """Downgrade database."""
    op.drop_index(op.f("ix_name_metadata_internal_id"), table_name="name_metadata")
    op.create_unique_constraint(
        "uq_name_metadata_internal_id", "name_metadata", ["internal_id"]
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""XML utils."""

from collections import defaultdict


def etree_to_dict(tree):
    """Convert an ElementTree to a dictionary."""
    tag = tree.tag.split(":")[-1]  # strip namespace
    d = {tag: {} if tree.attrib else None}
    children = list(tree)
    if children:
        dd = defaultdict(list)
        for dc in map(etree_to_dict, children):
            for k, v in dc.items():
                dd[k].append(v)
        d = {tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
    if tree.attrib:
        d[tag].update(("@" + k, v) for k, v in tree.attrib.items())
    if tree.text:
        text = tree.text.strip()
        if children or tree.attrib:
            if text:
                d[tag]["#text"] = text
        else:
            d[tag] = text
    return d



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Pytest configuration.

See https://pytest-invenio.readthedocs.io/ for documentation on which test
fixtures are available.
"""

import pytest
from invenio_records_resources.proxies import current_service_registry


@pytest.fixture(scope="function")
def affiliation_full_data():
    """Full affiliation data."""
    return {
        "acronym": "CERN",
        "id": "01ggx4157",
        "identifiers": [{"identifier": "03yrm5c26", "scheme": "ror"}],
        "name": "Test affiliation",
        "title": {"en": "Test affiliation", "es": "Afiliacion de test"},
        "country": "CH",
        "country_name": "Switzerland",
        "location_name": "Geneva",
        "status": "active",
        "types": ["facility", "funder"],
    }


@pytest.fixture(scope="function")
def affiliation_openaire_data():
    """Full affiliation data."""
    return {
        "acronym": "CERN",
        "id": "01ggx4157",
        "identifiers": [{"identifier": "999988133", "scheme": "pic"}],
        "name": "Test affiliation",
        "title": {"en": "Test affiliation", "es": "Afiliacion de test"},
        "country": "CH",
        "country_name": "Switzerland",
        "location_name": "Geneva",
        "status": "active",
        "types": ["facility", "funder"],
    }


@pytest.fixture(scope="function")
def openaire_affiliation_full_data():
    """Full OpenAIRE affiliation data."""
    return {
        "id": "01ggx4157",
        "identifiers": [{"identifier": "999988133", "scheme": "pic"}],
    }


@pytest.fixture(scope="module")
def service():
    """Affiliations service object."""
    return current_service_registry.get("affiliations")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
# Copyright (C) 2024 KTH Royal Institute of Technology.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
"""Test VocabularyConfig."""

import pytest
from invenio_records_resources.proxies import current_service_registry

from invenio_vocabularies.contrib.affiliations.datastreams import (
    DATASTREAM_CONFIG as affiliations_ds_config,
)
from invenio_vocabularies.contrib.awards.datastreams import (
    DATASTREAM_CONFIG as awards_ds_config,
)
from invenio_vocabularies.contrib.funders.datastreams import (
    DATASTREAM_CONFIG as funders_ds_config,
)
from invenio_vocabularies.contrib.names.datastreams import (
    DATASTREAM_CONFIG as names_ds_config,
)
from invenio_vocabularies.factories import (
    AffiliationsVocabularyConfig,
    AwardsVocabularyConfig,
    FundersVocabularyConfig,
    NamesVocabularyConfig,
)


@pytest.mark.parametrize(
    "conf, ds_config, service_type",
    [
        (AwardsVocabularyConfig(), awards_ds_config, "awards-service"),
        (NamesVocabularyConfig(), names_ds_config, "names-service"),
        (FundersVocabularyConfig(), funders_ds_config, "funders-service"),
        (
            AffiliationsVocabularyConfig(),
            affiliations_ds_config,
            "affiliations-service",
        ),
    ],
)
def test_vocabulary_config(conf, ds_config, service_type, app):
    """Test VocabularyConfig."""
    config = conf.get_config()

    # Compare readers
    assert config["readers"] == ds_config["readers"]

    # Compare writers, excluding 'identity'
    for config_writer, expected_writer in zip(config["writers"], ds_config["writers"]):
        assert config_writer["type"] == expected_writer["type"]
        assert str(config_writer["args"]) == str(expected_writer["args"])
        if config["writers"][0]["type"] == "async":
            assert config["writers"][0]["args"]["writer"]["type"] == service_type
        else:
            assert config["writers"][0]["type"] == service_type


def test_names_service(app):
    """Test service retrieval for names."""
    names_conf = NamesVocabularyConfig()
    service = names_conf.get_service()
    assert service == current_service_registry.get("names")


@pytest.mark.parametrize(
    "config_class, expected_error",
    [
        (FundersVocabularyConfig, NotImplementedError),
        (AwardsVocabularyConfig, NotImplementedError),
    ],
)
def test_service_not_implemented(config_class, expected_error, app):
    """Test services that are not implemented."""
    with pytest.raises(expected_error):
        conf = config_class()
        conf.get_service()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

import io

import pytest
from rdflib import Graph

from invenio_vocabularies.contrib.subjects.euroscivoc.datastreams import (
    EuroSciVocSubjectsTransformer,
)
from invenio_vocabularies.datastreams.datastreams import StreamEntry
from invenio_vocabularies.datastreams.readers import RDFReader

XML_DATA_PREF_LABEL = bytes(
    """<?xml version="1.0" encoding="UTF-8"?>
<rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#" xmlns:dcterms="http://purl.org/dc/terms/" xmlns:owl="http://www.w3.org/2002/07/owl#" xmlns:skos="http://www.w3.org/2004/02/skos/core#" xmlns:ns5="http://publications.europa.eu/ontology/euvoc#" xmlns:ns6="http://www.w3.org/2008/05/skos-xl#" xmlns:dc="http://purl.org/dc/elements/1.1/">
    <rdf:Description rdf:about="http://data.europa.eu/8mn/euroscivoc/87ff3577-527a-4a40-9c76-2f9d3075e2ba">
        <rdf:type rdf:resource="http://www.w3.org/2004/02/skos/core#Concept"/>
        <dcterms:created rdf:datatype="http://www.w3.org/2001/XMLSchema#date">2019-12-02</dcterms:created>
        <dcterms:modified rdf:datatype="http://www.w3.org/2001/XMLSchema#date">2019-12-02</dcterms:modified>
        <owl:versionInfo>1.1.0</owl:versionInfo>
        <skos:inScheme rdf:resource="http://data.europa.eu/8mn/euroscivoc/40c0f173-baa3-48a3-9fe6-d6e8fb366a00"/>
        <skos:prefLabel xml:lang="it">radio satellitare</skos:prefLabel>
        <skos:prefLabel xml:lang="pl">radio satelitarne</skos:prefLabel>
        <skos:prefLabel xml:lang="fr">radio satellite</skos:prefLabel>
        <skos:prefLabel xml:lang="es">radio por satélite</skos:prefLabel>
        <skos:prefLabel xml:lang="de">Satellitenfunk</skos:prefLabel>
        <skos:prefLabel xml:lang="en">satellite radio</skos:prefLabel>
        <ns5:startDate rdf:datatype="http://www.w3.org/2001/XMLSchema#date">2019-12-02</ns5:startDate>
        <skos:notation>87ff3577-527a-4a40-9c76-2f9d3075e2ba</skos:notation>
        <skos:notation>1717</skos:notation>
        <ns5:status rdf:resource="http://publications.europa.eu/resource/authority/concept-status/CURRENT"/>
        <ns5:xlNotation rdf:resource="http://data.europa.eu/8mn/euroscivoc/7f282340-9125-4b0d-aceb-23389311e306"/>
        <ns5:xlNotation rdf:resource="http://data.europa.eu/8mn/euroscivoc/notation_6014c8c4f809ee00ced00312669d98ad"/>
        <skos:altLabel xml:lang="en">broadcastingsatellite service</skos:altLabel>
        <skos:altLabel xml:lang="en">satellite radio system</skos:altLabel>
        <skos:broader rdf:resource="http://data.europa.eu/8mn/euroscivoc/d913bd42-e79c-46a7-8714-14f2a6a0d82f"/>
        <dc:identifier>87ff3577-527a-4a40-9c76-2f9d3075e2ba</dc:identifier>
        <owl:deprecated rdf:datatype="http://www.w3.org/2001/XMLSchema#boolean">false</owl:deprecated>
        </rdf:Description>
        <rdf:Description rdf:about="http://data.europa.eu/8mn/euroscivoc/d913bd42-e79c-46a7-8714-14f2a6a0d82f">
        <skos:narrower rdf:resource="http://data.europa.eu/8mn/euroscivoc/87ff3577-527a-4a40-9c76-2f9d3075e2ba"/>
    </rdf:Description>
     <rdf:Description rdf:about="http://data.europa.eu/8mn/euroscivoc/87ff3577-527a-4a40-9c76-2f9d3075e2ba">
            <skos:broader rdf:resource="http://data.europa.eu/8mn/euroscivoc/d913bd42-e79c-46a7-8714-14f2a6a0d82f"/>
            </rdf:Description>
            <rdf:Description rdf:about="http://data.europa.eu/8mn/euroscivoc/8d83b645-355f-4cf1-abf3-ce4cd3172c34">
            <skos:broader rdf:resource="http://data.europa.eu/8mn/euroscivoc/d913bd42-e79c-46a7-8714-14f2a6a0d82f"/>
            </rdf:Description>
            <rdf:Description rdf:about="http://data.europa.eu/8mn/euroscivoc/d913bd42-e79c-46a7-8714-14f2a6a0d82f">
            <rdf:type rdf:resource="http://www.w3.org/2004/02/skos/core#Concept"/>
            <dcterms:created rdf:datatype="http://www.w3.org/2001/XMLSchema#date">2019-12-02</dcterms:created>
            <dcterms:modified rdf:datatype="http://www.w3.org/2001/XMLSchema#date">2023-03-10</dcterms:modified>
            <owl:versionInfo>1.3.4</owl:versionInfo>
            <skos:inScheme rdf:resource="http://data.europa.eu/8mn/euroscivoc/40c0f173-baa3-48a3-9fe6-d6e8fb366a00"/>
            <ns5:startDate rdf:datatype="http://www.w3.org/2001/XMLSchema#date">2019-12-02</ns5:startDate>
            <skos:notation>1225</skos:notation>
            <skos:notation>d913bd42-e79c-46a7-8714-14f2a6a0d82f</skos:notation>
            <ns5:status rdf:resource="http://publications.europa.eu/resource/authority/concept-status/CURRENT"/>
            <ns5:xlNotation rdf:resource="http://data.europa.eu/8mn/euroscivoc/notation_1391bd2069b8dde8f6394b6a3b4241cb"/>
            <ns5:xlNotation rdf:resource="http://data.europa.eu/8mn/euroscivoc/1faa45c8-4f46-45d6-9d6f-447a2626125d"/>
            <skos:altLabel xml:lang="en">radio channel</skos:altLabel>
            <skos:altLabel xml:lang="en">wireless</skos:altLabel>
            <skos:broader rdf:resource="http://data.europa.eu/8mn/euroscivoc/1198b23a-f82f-4189-8778-d9a742430a0f"/>
            <dc:identifier>d913bd42-e79c-46a7-8714-14f2a6a0d82f</dc:identifier>
            <owl:deprecated rdf:datatype="http://www.w3.org/2001/XMLSchema#boolean">false</owl:deprecated>
        </rdf:Description>
</rdf:RDF>""",
    encoding="utf-8",
)


XML_DATA_ALT_LABEL = bytes(
    """<?xml version="1.0" encoding="UTF-8"?>
        <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#" xmlns:dcterms="http://purl.org/dc/terms/" xmlns:owl="http://www.w3.org/2002/07/owl#" xmlns:skos="http://www.w3.org/2004/02/skos/core#" xmlns:ns5="http://publications.europa.eu/ontology/euvoc#" xmlns:ns6="http://www.w3.org/2008/05/skos-xl#" xmlns:dc="http://purl.org/dc/elements/1.1/">
        <rdf:Description rdf:about="http://data.europa.eu/8mn/euroscivoc/87ff3577-527a-4a40-9c76-2f9d3075e2ba">
        <rdf:type rdf:resource="http://www.w3.org/2004/02/skos/core#Concept"/>
        <dcterms:created rdf:datatype="http://www.w3.org/2001/XMLSchema#date">2019-12-02</dcterms:created>
        <dcterms:modified rdf:datatype="http://www.w3.org/2001/XMLSchema#date">2019-12-02</dcterms:modified>
        <owl:versionInfo>1.1.0</owl:versionInfo>
        <skos:inScheme rdf:resource="http://data.europa.eu/8mn/euroscivoc/40c0f173-baa3-48a3-9fe6-d6e8fb366a00"/>
        <ns5:startDate rdf:datatype="http://www.w3.org/2001/XMLSchema#date">2019-12-02</ns5:startDate>
        <skos:notation>87ff3577-527a-4a40-9c76-2f9d3075e2ba</skos:notation>
        <skos:notation>1717</skos:notation>
        <ns5:status rdf:resource="http://publications.europa.eu/resource/authority/concept-status/CURRENT"/>
        <ns5:xlNotation rdf:resource="http://data.europa.eu/8mn/euroscivoc/7f282340-9125-4b0d-aceb-23389311e306"/>
        <ns5:xlNotation rdf:resource="http://data.europa.eu/8mn/euroscivoc/notation_6014c8c4f809ee00ced00312669d98ad"/>
        <skos:altLabel xml:lang="en">broadcastingsatellite service</skos:altLabel>
        <skos:altLabel xml:lang="en">satellite radio system</skos:altLabel>
        <skos:broader rdf:resource="http://data.europa.eu/8mn/euroscivoc/d913bd42-e79c-46a7-8714-14f2a6a0d82f"/>
        <dc:identifier>87ff3577-527a-4a40-9c76-2f9d3075e2ba</dc:identifier>
        <owl:deprecated rdf:datatype="http://www.w3.org/2001/XMLSchema#boolean">false</owl:deprecated>
        </rdf:Description>
        <rdf:Description rdf:about="http://data.europa.eu/8mn/euroscivoc/d913bd42-e79c-46a7-8714-14f2a6a0d82f">
        <skos:narrower rdf:resource="http://data.europa.eu/8mn/euroscivoc/87ff3577-527a-4a40-9c76-2f9d3075e2ba"/>
        </rdf:Description>
</rdf:RDF>""",
    encoding="utf-8",
)


@pytest.fixture(scope="module")
def expected_from_rdf_pref_label_with_parent():
    return [
        {
            "id": "euroscivoc:1717",
            "scheme": "EuroSciVoc",
            "subject": "Satellite radio",
            "title": {
                "it": "Radio satellitare",
                "pl": "Radio satelitarne",
                "fr": "Radio satellite",
                "es": "Radio por satélite",
                "de": "Satellitenfunk",
                "en": "Satellite radio",
            },
            "props": {"parents": "euroscivoc:1225"},
            "identifiers": [
                {
                    "scheme": "url",
                    "identifier": "http://data.europa.eu/8mn/euroscivoc/87ff3577-527a-4a40-9c76-2f9d3075e2ba",
                }
            ],
        },
        {
            "id": "euroscivoc:1225",
            "scheme": "EuroSciVoc",
            "subject": "Radio channel",
            "title": {"en": "Radio channel"},
            "props": {},
            "identifiers": [
                {
                    "scheme": "url",
                    "identifier": "http://data.europa.eu/8mn/euroscivoc/d913bd42-e79c-46a7-8714-14f2a6a0d82f",
                }
            ],
        },
    ]


@pytest.fixture(scope="module")
def expected_from_rdf_alt_label_without_parent():
    return {
        "id": "euroscivoc:1717",
        "scheme": "EuroSciVoc",
        "subject": "Broadcastingsatellite service",
        "title": {"en": "Broadcastingsatellite service"},
        "props": {},
        "identifiers": [
            {
                "scheme": "url",
                "identifier": "http://data.europa.eu/8mn/euroscivoc/87ff3577-527a-4a40-9c76-2f9d3075e2ba",
            }
        ],
    }


def test_euroscivoc_subjects_transformer_pref_label(
    expected_from_rdf_pref_label_with_parent,
):
    reader = RDFReader()
    rdf_data = io.BytesIO(XML_DATA_PREF_LABEL)
    rdf_graph = Graph()
    rdf_graph.parse(rdf_data, format="xml")
    stream_entries = list(reader._iter(rdf_graph))
    assert len(stream_entries) > 0
    transformer = EuroSciVocSubjectsTransformer()
    result = []
    for entry in stream_entries:
        entry = transformer.apply(StreamEntry(entry)).entry
        result.append(entry)
    assert expected_from_rdf_pref_label_with_parent == result


def test_euroscivoc_subjects_transformer_alt_label(
    expected_from_rdf_alt_label_without_parent,
):
    reader = RDFReader()
    rdf_data = io.BytesIO(XML_DATA_ALT_LABEL)
    rdf_graph = Graph()
    rdf_graph.parse(rdf_data, format="xml")
    stream_entries = list(reader._iter(rdf_graph))
    assert len(stream_entries) > 0
    transformer = EuroSciVocSubjectsTransformer()
    for entry in stream_entries:
        result = transformer.apply(StreamEntry(entry))
        assert expected_from_rdf_alt_label_without_parent == result.entry



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Datastreams module."""

from .datastreams import DataStream, StreamEntry
from .factories import DataStreamFactory

__all__ = (
    "DataStream",
    "DataStreamFactory",
    "StreamEntry",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Funders JSONSchema tests."""

import pytest
from jsonschema.exceptions import ValidationError

from invenio_vocabularies.contrib.funders.api import Funder


@pytest.fixture(scope="module")
def schema():
    """Returns the schema location."""
    return "local://funders/funder-v1.0.0.json"


def validates(data):
    """Validates funder data."""
    Funder(data).validate()

    return True


def fails(data):
    """Validates funder data."""
    pytest.raises(ValidationError, validates, data)
    return True


def test_valid_full(appctx, schema):
    data = {
        "$schema": schema,
        "name": "Test funder",
        "identifiers": [
            {
                "identifier": "0000000123456X",
                "scheme": "isni",
            },
            {
                "identifier": "grid.1234.5",
                "scheme": "grid",
            },
        ],
        "country": "CH",
    }

    assert validates(data)


def test_valid_empty(appctx, schema):
    # check there are no requirements at JSONSchema level
    data = {"$schema": schema}

    assert validates(data)


# name and country are defined by the funder schema
# the rest are inherited and should be tested elsewhere


def test_fails_name(appctx, schema):
    data = {"$schema": schema, "name": 1}

    assert fails(data)


def test_fails_country(appctx, schema):
    data = {"$schema": schema, "country": 1}

    assert fails(data)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
# Copyright (C) 2024 University of Münster.
#
# Invenio-Vocabularies is free software; you can redistringibute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Resources config."""

import marshmallow as ma
from flask_resources import (
    BaseListSchema,
    HTTPJSONException,
    JSONSerializer,
    MarshmallowSerializer,
    ResourceConfig,
    ResponseHandler,
    create_error_handler,
)
from invenio_records_resources.resources import (
    RecordResource,
    RecordResourceConfig,
    SearchRequestArgsSchema,
)
from invenio_records_resources.resources.errors import ErrorHandlersMixin
from invenio_records_resources.resources.records.args import SearchRequestArgsSchema
from invenio_records_resources.resources.records.headers import etag_headers
from invenio_records_resources.services.base.config import ConfiguratorMixin

from .serializer import VocabularyL10NItemSchema


class VocabularySearchRequestArgsSchema(SearchRequestArgsSchema):
    """Vocabularies search request parameters."""

    tags = ma.fields.Str()
    active = ma.fields.Boolean()
    status = ma.fields.Boolean()


class VocabulariesResourceConfig(RecordResourceConfig):
    """Vocabulary resource configuration."""

    blueprint_name = "vocabularies"
    url_prefix = "/vocabularies"
    routes = {
        "list": "/<type>",
        "item": "/<type>/<pid_value>",
        "tasks": "/tasks",
    }

    request_view_args = {
        "pid_value": ma.fields.Str(),
        "type": ma.fields.Str(required=True),
    }

    request_search_args = VocabularySearchRequestArgsSchema

    response_handlers = {
        "application/json": ResponseHandler(JSONSerializer(), headers=etag_headers),
        "application/vnd.inveniordm.v1+json": ResponseHandler(
            MarshmallowSerializer(
                format_serializer_cls=JSONSerializer,
                object_schema_cls=VocabularyL10NItemSchema,
                list_schema_cls=BaseListSchema,
            ),
            headers=etag_headers,
        ),
    }


class VocabularyTypeResourceConfig(ResourceConfig, ConfiguratorMixin):
    """Vocabulary list resource config."""

    # /vocabulary/vocabulary_id
    # Blueprint configuration
    blueprint_name = "vocabulary_types"
    url_prefix = "/vocabularies"

    routes = {"list": "/"}

    # Request parsing
    request_read_args = {}
    request_view_args = {
        "pid_value": ma.fields.String,
        "type": ma.fields.String,
    }
    request_search_args = VocabularySearchRequestArgsSchema

    error_handlers = {
        **ErrorHandlersMixin.error_handlers,
        # TODO: Add custom error handlers here
    }
    response_handlers = {
        "application/json": ResponseHandler(JSONSerializer(), headers=etag_headers),
        "application/vnd.inveniordm.v1+json": ResponseHandler(
            MarshmallowSerializer(
                format_serializer_cls=JSONSerializer,
                object_schema_cls=VocabularyL10NItemSchema,
                list_schema_cls=BaseListSchema,
            ),
            headers=etag_headers,
        ),
    }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""BODC subjects datastreams, readers, transformers, and writers."""

from invenio_vocabularies.datastreams.errors import TransformerError
from invenio_vocabularies.datastreams.readers import RDFReader
from invenio_vocabularies.datastreams.transformers import RDFTransformer

from ..config import bodc_puv_file_url

# Available with the "rdf" extra
try:
    import rdflib
except ImportError:
    rdflib = None


class BODCPUVSubjectsTransformer(RDFTransformer):
    """
    Transformer class to convert BODC-PUV RDF data to a dictionary format.

    Input:
        - Relevant fields:
            - `skos:notation`: Primary identifier for the concept.
            - `skos:prefLabel`: Preferred labels with language codes.
            - `skos:altLabel`: Alternative labels (optional).
            - `skos:definition`: Definitions of the concept.
            - `owl:deprecated`: Boolean flag indicating if the concept is deprecated.

    Output:
        - A dictionary with the following structure:
            {
                "id": "SDN:P01::SAGEMSFM",  # BODC-specific parameter ID (skos:notation).
                "scheme": "BODC-PUV",  # The scheme name indicating this is a BODC Parameter Usage Vocabulary concept.
                "subject": "AMSSedAge",  # The alternative label (skos:altLabel), if available, or None.
                "title": {
                    "en": "14C age of Foraminiferida"  # English preferred label (skos:prefLabel).
                },
                "props": {
                    "definitions": "Accelerated mass spectrometry on picked tests",  # Definition of subject (skos:definition).
                },
                "identifiers": [
                    {
                        "scheme": "url",  # Type of identifier (URL).
                        "identifier": "http://vocab.nerc.ac.uk/collection/P01/current/SAGEMSFM"  # URI of the concept.
                    }
                ]
            }
    """

    def _get_subject_data(self, rdf_graph, subject):
        """Fetch all triples for a subject and organize them into a dictionary."""
        data = {}
        for predicate, obj in rdf_graph.predicate_objects(subject=subject):
            predicate_name = str(predicate)
            if predicate_name not in data:
                data[predicate_name] = []
            data[predicate_name].append(obj)
        return data

    def _transform_entry(self, subject, rdf_graph):
        """Transform an entry to the required dictionary format."""
        labels = self._get_labels(subject, rdf_graph)
        subject_data = self._get_subject_data(rdf_graph, subject)
        deprecated = subject_data.get(str(rdflib.namespace.OWL.deprecated), [False])
        if deprecated and str(deprecated[0]).lower() == "true":
            return None  # Skip deprecated subjects

        notation = subject_data.get(str(self.skos_core.notation), [])
        if notation:
            id = str(notation[0])
        else:
            raise TransformerError(f"No id found for: {subject}")

        alt_labels = [obj for obj in subject_data.get(str(self.skos_core.altLabel), [])]
        subject_text = str(alt_labels[0]) if alt_labels else ""
        definition = str(subject_data.get(str(self.skos_core.definition), [None])[0])

        return {
            "id": id,
            "scheme": "BODC-PUV",
            "subject": subject_text,
            "title": labels,
            "props": {"definition": definition} if definition else {},
            "identifiers": self._get_identifiers(subject),
        }


# Configuration for datastream

VOCABULARIES_DATASTREAM_TRANSFORMERS = {"bodc-transformer": BODCPUVSubjectsTransformer}

DATASTREAM_CONFIG = {
    "readers": [
        {
            "type": "http",
            "args": {
                "origin": bodc_puv_file_url,
            },
        },
        {"type": "rdf"},
    ],
    "transformers": [{"type": "bodc-transformer"}],
    "writers": [{"args": {"writer": {"type": "subjects-service"}}, "type": "async"}],
}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2024 CERN.
# Copyright (C) 2024 California Institute of Technology.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Affiliations schema."""

from functools import partial

from invenio_i18n import lazy_gettext as _
from marshmallow import fields, validate
from marshmallow_utils.fields import IdentifierSet, SanitizedUnicode
from marshmallow_utils.schemas import IdentifierSchema

from ...services.schema import (
    BaseVocabularySchema,
    ContribVocabularyRelationSchema,
    ModePIDFieldVocabularyMixin,
)
from .config import affiliation_schemes


class AffiliationSchema(BaseVocabularySchema, ModePIDFieldVocabularyMixin):
    """Service schema for affiliations."""

    acronym = SanitizedUnicode()
    identifiers = IdentifierSet(
        fields.Nested(
            partial(
                IdentifierSchema,
                allowed_schemes=affiliation_schemes,
                identifier_required=False,
            )
        )
    )
    name = SanitizedUnicode(
        required=True, validate=validate.Length(min=1, error=_("Name cannot be blank."))
    )
    country = SanitizedUnicode()
    country_name = SanitizedUnicode()
    location_name = SanitizedUnicode()
    id = SanitizedUnicode(
        validate=validate.Length(min=1, error=_("PID cannot be blank."))
    )

    acronym = SanitizedUnicode()
    aliases = fields.List(SanitizedUnicode())
    status = SanitizedUnicode()
    types = fields.List(SanitizedUnicode())


class AffiliationRelationSchema(ContribVocabularyRelationSchema):
    """Schema to define an optional affialiation relation in another schema."""

    ftf_name = "name"
    parent_field_name = "affiliations"
    name = SanitizedUnicode()
    identifiers = IdentifierSet(
        fields.Nested(
            partial(
                IdentifierSchema,
                allowed_schemes=affiliation_schemes,
                identifier_required=False,
            )
        ),
        dump_only=True,
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 CERN.
# Copyright (C) 2024 California Institute of Technology.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Funders datastreams tests."""

from copy import deepcopy

import pytest
from invenio_access.permissions import system_identity

from invenio_vocabularies.contrib.common.ror.datastreams import RORTransformer
from invenio_vocabularies.contrib.funders.api import Funder
from invenio_vocabularies.contrib.funders.config import (
    funder_fundref_doi_prefix,
    funder_schemes,
)
from invenio_vocabularies.contrib.funders.datastreams import FundersServiceWriter
from invenio_vocabularies.datastreams import StreamEntry
from invenio_vocabularies.datastreams.errors import WriterError


@pytest.fixture(scope="module")
def expected_from_ror_json():
    return {
        "id": "05dxps055",
        "name": "California Institute of Technology",
        "title": {
            "en": "California Institute of Technology",
            "es": "Instituto de Tecnología de California",
        },
        "acronym": "CIT",
        "aliases": ["Caltech"],
        "country": "US",
        "country_name": "United States",
        "location_name": "Pasadena",
        "status": "active",
        "identifiers": [
            {"scheme": "ror", "identifier": "05dxps055"},
            {"scheme": "doi", "identifier": "10.13039/100006961"},
            {"scheme": "grid", "identifier": "grid.20861.3d"},
            {"scheme": "isni", "identifier": "0000 0001 0706 8890"},
        ],
        "types": ["education", "funder"],
    }


def test_ror_transformer(app, dict_ror_entry, expected_from_ror_json):
    transformer = RORTransformer(
        vocab_schemes=funder_schemes,
        funder_fundref_doi_prefix=funder_fundref_doi_prefix,
    )
    assert expected_from_ror_json == transformer.apply(dict_ror_entry).entry


def test_funders_service_writer_create(app, search_clear, funder_full_data):
    writer = FundersServiceWriter()
    funder_rec = writer.write(StreamEntry(funder_full_data))
    funder_dict = funder_rec.entry.to_dict()
    assert dict(funder_dict, **funder_full_data) == funder_dict

    # not-ideal cleanup
    funder_rec.entry._record.delete(force=True)


def test_funders_service_writer_duplicate(app, search_clear, funder_full_data):
    writer = FundersServiceWriter()
    funder_rec = writer.write(stream_entry=StreamEntry(funder_full_data))
    Funder.index.refresh()  # refresh index to make changes live
    with pytest.raises(WriterError) as err:
        writer.write(stream_entry=StreamEntry(funder_full_data))

    expected_error = [f"Vocabulary entry already exists: {funder_full_data}"]
    assert expected_error in err.value.args

    # not-ideal cleanup
    funder_rec.entry._record.delete(force=True)


def test_funders_service_writer_update_existing(
    app, search_clear, funder_full_data, service
):
    # create vocabulary
    writer = FundersServiceWriter(update=True)
    orig_funder_rec = writer.write(stream_entry=StreamEntry(funder_full_data))
    Funder.index.refresh()  # refresh index to make changes live
    # update vocabulary
    updated_funder = deepcopy(funder_full_data)
    updated_funder["name"] = "Updated Name"
    # check changes vocabulary
    _ = writer.write(stream_entry=StreamEntry(updated_funder))
    funder_rec = service.read(system_identity, orig_funder_rec.entry.id)
    funder_dict = funder_rec.to_dict()

    # needed while the writer resolves from ES
    assert _.entry.id == orig_funder_rec.entry.id
    assert dict(funder_dict, **updated_funder) == funder_dict

    # not-ideal cleanup
    funder_rec._record.delete(force=True)


def test_funders_service_writer_update_non_existing(
    app, search_clear, funder_full_data, service
):
    # vocabulary item not created, call update directly
    updated_funder = deepcopy(funder_full_data)
    updated_funder["name"] = "New name"
    # check changes vocabulary
    writer = FundersServiceWriter(update=True)
    funder_rec = writer.write(stream_entry=StreamEntry(updated_funder))
    funder_rec = service.read(system_identity, funder_rec.entry.id)
    funder_dict = funder_rec.to_dict()

    assert dict(funder_dict, **updated_funder) == funder_dict

    # not-ideal cleanup
    funder_rec._record.delete(force=True)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Names services."""

from invenio_pidstore.errors import PIDDoesNotExistError
from invenio_search.engine import dsl

from .names import record_type

NamesServiceConfig = record_type.service_config_cls


class NamesService(record_type.service_cls):
    """Name service."""

    def resolve(self, identity, id_, id_type, many=False):
        """Get the record with a given identifier.

        param id_: The identifier value.
        param id_type: The identifier type.
        param many: If True, return a list of records.
        """
        search_query = dsl.Q(
            "bool",
            minimum_should_match=1,
            must=[
                dsl.Q("term", identifiers__identifier=id_),
                dsl.Q("term", identifiers__scheme=id_type),
            ],
        )

        # max_records = 1, we assume there cannot be duplicates
        # the loading process needs to make sure of that
        if many:
            results = self._read_many(identity, search_query)
        else:
            results = self._read_many(identity, search_query, max_records=1)

        # cant use the results_item because it returns dicts instead of records
        total = results.hits.total["value"]
        if total == 0:
            # Not a PID but treated as such
            raise PIDDoesNotExistError(pid_type=id_type, pid_value=id_)
        if many:
            for result in results:
                record = self.record_cls.loads(result.to_dict())
                self.require_permission(identity, "read", record=record)
            return self.result_list(self, identity, results)
        else:
            record = self.record_cls.loads(results[0].to_dict())
            self.require_permission(identity, "read", record=record)

            return self.result_item(
                self,
                identity,
                record,
                links_tpl=self.links_item_tpl,
            )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Module tests."""

from flask import Flask

from invenio_vocabularies import InvenioVocabularies


def test_version():
    """Test version import."""
    from invenio_vocabularies import __version__

    assert __version__


def test_init():
    """Test extension initialization."""
    app = Flask("testapp")
    ext = InvenioVocabularies(app)
    assert "invenio-vocabularies" in app.extensions

    app = Flask("testapp")
    ext = InvenioVocabularies()
    assert "invenio-vocabularies" not in app.extensions
    ext.init_app(app)
    assert "invenio-vocabularies" in app.extensions



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Names JSONSchema tests."""

import pytest
from jsonschema.exceptions import ValidationError

from invenio_vocabularies.contrib.names.api import Name


@pytest.fixture(scope="module")
def schema():
    """Returns the schema location."""
    return "local://names/name-v1.0.0.json"


def validates(data):
    """Validates names data."""
    Name(data).validate()

    return True


def fails(data):
    """Validates names data."""
    pytest.raises(ValidationError, validates, data)
    return True


def test_valid_full(appctx, schema):
    data = {
        "$schema": schema,
        "name": "Doe, John",
        "given_name": "John",
        "family_name": "Doe",
        "identifiers": [{"identifier": "0000-0001-8135-3489", "scheme": "orcid"}],
        "affiliations": [{"id": "cern"}, {"name": "CustomORG"}],
    }
    assert validates(data)


def test_valid_empty(appctx, schema):
    # check there are no requirements at JSONSchema level
    data = {"$schema": schema}

    assert validates(data)


# only acronym and name are defined by the affiliation schema
# the rest are inherited and should be tested elsewhere


def test_fails_name_identifiers(appctx, schema):
    # string
    data = {"$schema": schema, "identifiers": "0000-0001-8135-3489"}

    assert fails(data)

    # dict
    data = {
        "$schema": schema,
        "identifiers": {"identifier": "0000-0001-8135-3489", "scheme": "orcid"},
    }

    assert fails(data)


def test_fails_name_affiliations(appctx, schema):
    # string, comma separated list
    data = {"$schema": schema, "affiliations": "cern, CustomORG"}

    assert fails(data)

    # dict
    data = {"$schema": schema, "affiliations": {"id": "cern", "name": "CustomORG"}}

    assert fails(data)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 CERN.
# Copyright (C) 2024 California Institute of Technology.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Affiliations datastreams, transformers, writers and readers."""

from copy import deepcopy

import pycountry
from flask import current_app

from ...datastreams import StreamEntry
from ...datastreams.errors import TransformerError
from ...datastreams.transformers import BaseTransformer
from ...datastreams.writers import ServiceWriter
from ..affiliations.config import affiliation_edmo_country_mappings
from ..common.ror.datastreams import RORTransformer


class AffiliationsServiceWriter(ServiceWriter):
    """Affiliations service writer."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        service_or_name = kwargs.pop("service_or_name", "affiliations")
        super().__init__(service_or_name=service_or_name, *args, **kwargs)

    def _entry_id(self, entry):
        """Get the id from an entry."""
        return entry["id"]


class AffiliationsRORTransformer(RORTransformer):
    """Affiliations ROR Transformer."""

    def __init__(
        self, *args, vocab_schemes=None, funder_fundref_doi_prefix=None, **kwargs
    ):
        """Constructor."""
        if vocab_schemes is None:
            vocab_schemes = current_app.config.get("VOCABULARIES_AFFILIATION_SCHEMES")
        super().__init__(
            *args,
            vocab_schemes=vocab_schemes,
            funder_fundref_doi_prefix=funder_fundref_doi_prefix,
            **kwargs,
        )


class OpenAIREOrganizationTransformer(BaseTransformer):
    """OpenAIRE Organization Transformer."""

    def apply(self, stream_entry, **kwargs):
        """Applies the transformation to the stream entry."""
        record = stream_entry.entry

        if "id" not in record:
            raise TransformerError([f"No id for: {record}"])

        if not record["id"].startswith("openorgs____::"):
            raise TransformerError([f"Not valid OpenAIRE OpenOrgs id for: {record}"])

        if "pid" not in record:
            raise TransformerError([f"No alternative identifiers for: {record}"])

        organization = {}

        for pid in record["pid"]:
            if pid["scheme"] == "ROR":
                organization["id"] = pid["value"].removeprefix("https://ror.org/")
            elif pid["scheme"] == "PIC":
                organization["identifiers"] = [
                    {
                        "scheme": "pic",
                        "identifier": pid["value"],
                    }
                ]

        stream_entry.entry = organization
        return stream_entry


class OpenAIREAffiliationsServiceWriter(ServiceWriter):
    """OpenAIRE Affiliations service writer."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        service_or_name = kwargs.pop("service_or_name", "affiliations")
        # Here we only update and we do not insert, since OpenAIRE data is used to augment existing affiliations
        # (with PIC identifiers) and is not used to create new affiliations.
        super().__init__(
            service_or_name=service_or_name, insert=False, update=True, *args, **kwargs
        )

    def _entry_id(self, entry):
        """Get the id from an entry."""
        return entry["id"]

    def _do_update(self, entry):
        vocab_id = self._entry_id(entry)
        current = self._resolve(vocab_id)
        updated = deepcopy(current.to_dict())

        if "identifiers" in entry:
            # For each new identifier
            for new_identifier in entry["identifiers"]:
                # Either find an existing identifier with the same scheme and update the "identifier" value
                for existing_identifier in updated["identifiers"]:
                    if existing_identifier["scheme"] == new_identifier["scheme"]:
                        existing_identifier["identifier"] = new_identifier["identifier"]
                        break
                # Or add the new identifier to the list of identifiers
                else:
                    updated["identifiers"].append(new_identifier)

        return StreamEntry(self._service.update(self._identity, vocab_id, updated))


class EDMOOrganizationTransformer(BaseTransformer):
    """Transformer class to convert EDMO RDF data to a dictionary format."""

    def apply(self, stream_entry, **kwargs):
        """Applies the transformation to the stream entry."""
        record = stream_entry.entry
        edmo_uri = record["org"]["value"]
        id_ = "edmo:" + edmo_uri.split("/")[-1]
        name = record["name"]["value"]
        alt_name = record.get("altName", {}).get("value")
        country_name = record.get("countryName", {}).get("value")
        locality = record.get("locality", {}).get("value")
        deprecated = record["deprecated"]["value"]

        # Fix organizations with the old country name "Cape Verde".
        # "Cabo Verde" is the new official name also used by ROR, e.g. https://api.ror.org/v2/organizations/001fphc23
        if country_name in affiliation_edmo_country_mappings:
            country_name = affiliation_edmo_country_mappings[country_name]

        # Logic to convert a country name to a 2 letters country code.
        country = None
        if country_name:
            country_dict = pycountry.countries.get(name=country_name)
            if country_dict:
                country = country_dict.alpha_2
            else:
                raise TransformerError([f"No alpha_2 country found for: {record}"])

        # Mandatory fields
        organization = {
            "id": id_,
            "identifiers": [
                {
                    "scheme": "edmo",
                    "identifier": edmo_uri,
                }
            ],
            "name": name,
            "title": {
                "en": name,
            },
        }

        # Optional fields
        if alt_name:
            organization["acronym"] = alt_name

        if country_name:
            organization["country_name"] = country_name

        if country:
            organization["country"] = country

        if locality:
            organization["location_name"] = locality

        stream_entry.entry = organization
        return stream_entry


VOCABULARIES_DATASTREAM_READERS = {}
"""Affiliations datastream readers."""

VOCABULARIES_DATASTREAM_WRITERS = {
    "affiliations-service": AffiliationsServiceWriter,
    "openaire-affiliations-service": OpenAIREAffiliationsServiceWriter,
}
"""Affiliations datastream writers."""

VOCABULARIES_DATASTREAM_TRANSFORMERS = {
    "ror-affiliations": AffiliationsRORTransformer,
    "openaire-organization": OpenAIREOrganizationTransformer,
    "edmo-organization": EDMOOrganizationTransformer,
}
"""Affiliations datastream transformers."""


DATASTREAM_CONFIG = {
    "readers": [
        {
            "type": "zip",
            "args": {
                "regex": "_schema_v2\\.json$",
            },
        },
        {"type": "json"},
    ],
    "transformers": [
        {
            "type": "ror-affiliations",
        },
    ],
    "writers": [
        {
            "type": "async",
            "args": {
                "writer": {
                    "type": "affiliations-service",
                }
            },
        }
    ],
}
"""Data Stream configuration.

An origin is required for the reader.
"""

DATASTREAM_CONFIG_OPENAIRE = {
    "readers": [
        {"type": "openaire-http", "args": {"tar_href": "/organization.tar"}},
        {
            "type": "tar",
            "args": {
                "regex": "\\.json.gz$",
                "mode": "r",
            },
        },
        {"type": "gzip"},
        {"type": "jsonl"},
    ],
    "transformers": [
        {
            "type": "openaire-organization",
        },
    ],
    "writers": [
        {
            "type": "async",
            "args": {
                "writer": {
                    "type": "openaire-affiliations-service",
                }
            },
        }
    ],
}
"""Alternative Data Stream configuration for OpenAIRE Affiliations."""

DATASTREAM_CONFIG_EDMO = {
    "readers": [
        {
            "type": "sparql",
            "args": {
                "origin": "https://edmo.seadatanet.org/sparql/sparql",
                "query": """
                    SELECT ?org ?name ?altName ?countryName ?locality ?deprecated
                    WHERE {
                        ?org a <http://www.w3.org/ns/org#Organization> .
                        ?org <http://www.w3.org/ns/org#name> ?name .
                        OPTIONAL { ?org <http://www.w3.org/2004/02/skos/core#altName> ?altName } .
                        OPTIONAL { ?org <http://www.w3.org/2006/vcard/ns#country-name> ?countryName } .
                        OPTIONAL { ?org <http://www.w3.org/2006/vcard/ns#locality> ?locality } .
                        OPTIONAL { ?org <http://www.w3.org/2002/07/owl#deprecated> ?deprecated } .
                        FILTER (!?deprecated)
                    }
                    """,
            },
        }
    ],
    "transformers": [
        {
            "type": "edmo-organization",
        },
    ],
    "writers": [
        {
            "type": "async",
            "args": {
                "writer": {
                    "type": "affiliations-service",
                }
            },
        },
    ],
}
"""Alternative Data Stream configuration for EDMO Affiliations."""
